import logging
import requests
import os
import re
import fnmatch
from typing import Optional, Literal
from gi.repository import Adw, Gio
from urllib.parse import urlsplit

from ..lib.utils import get_file_hash
from ..lib import json_config
from ..lib.ini_config import Config
from ..providers.AppImageProvider import AppImageProvider, AppImageListElement
from .Models import DownloadInterruptedException

from .UpdateManager import UpdateManager
from .StaticFileUpdater import StaticFileUpdater
from ..components.AdwEntryRowDefault import AdwEntryRowDefault

class GithubUpdater(UpdateManager):
    handles_embedded = 'gh-releases-zsync|'
    staticfile_manager: Optional[StaticFileUpdater]
    label = 'Github'
    name = 'GithubUpdater'

    def __init__(self, el, embedded=False) -> None:
        super().__init__(embedded=embedded, el=el)
        self.staticfile_manager = None
        self.repo_url_row = None
        self.repo_filename_row = None
        self.allow_prereleases_row = None

    def migrate_v2(self):
        app_config = json_config.read_config_for_app(self.el)
        config = None

        if 'update_manager_config' in app_config:
            old_config = app_config.get('update_manager_config', {})
            url_data = urlsplit(old_config.get('repo_url', '')).path.split('/')
            repo = ''

            if len(url_data) > 2:
                repo = '/'.join([url_data[1], url_data[2]])

            config = {
                'allow_prereleases':  old_config.get('allow_prereleases', False),
                'repo': repo,
                'repo_filename': old_config.get('repo_filename', ''),
            }
        elif 'update_url' in app_config:
            url_data = self.get_url_data(app_config['update_url'])

            if url_data:
                config = {
                    'allow_prereleases': False,
                    'repo': '/'.join([url_data['username'], url_data['repo']]),
                    'repo_filename': url_data['filename'],
                }
        
        if config:
            Config.set_app_update_config(self.el, self, config)

    def get_embedded_data(self):
        # Format gh-releases-zsync|probono|AppImages|latest|Subsurface-*x86_64.AppImage.zsync
        # https://github.com/AppImage/AppImageSpec/blob/master/draft.md#github-releases
        if not self.embedded:
            return None

        url = self.embedded

        tag_name = '*'
        items = url.split('|')

        if len(items) != 5:
            return None

        return {
            'username': items[1],
            'repo': items[2],
            'release': items[3],
            'filename': items[4],
            'tag_name': tag_name
        }
    
    def get_url_data(self, url: str):
        # Format gh-releases-zsync|probono|AppImages|latest|Subsurface-*x86_64.AppImage.zsync
        # https://github.com/AppImage/AppImageSpec/blob/master/draft.md#github-releases

        tag_name = '*'
        if url.startswith('https://'):
            logging.debug(f'GithubUpdater: found http url, trying to detect github data')
            urldata = urlsplit(url)

            if urldata.netloc != 'github.com':
                return None

            paths = urldata.path.split('/')

            if len(paths) < 7:
                return None

            if paths[3] != 'releases' or paths[4] != 'download':
                return None

            rel_name = 'latest'
            tag_name = paths[5]

            url = f'|{paths[1]}|{paths[2]}|{rel_name}|{paths[6]}'
            logging.debug(f'GithubUpdater: generated appimages-like update string "{url}"')

        items = url.split('|')

        if len(items) != 5:
            return None

        return {
            'username': items[1],
            'repo': items[2],
            'release': items[3],
            'filename': items[4],
            'tag_name': tag_name
        }

    def does_allow_prereleases(self):
        allow_prereleases = False

        if self.embedded:
            embedded_data = self.get_embedded_data()
            if embedded_data:
                allow_prereleases = embedded_data['release'] in ['latest-pre', 'latest-all']
        else:
            allow_prereleases = self.get_config().getboolean('allow_prereleases', False)

        return allow_prereleases

    def download(self, status_update_cb) -> tuple[str, str]:
        target_asset = self.fetch_target_asset()
        if not target_asset:
            raise Exception(f'Missing target_asset for {self.name} instance')

        dwnl = target_asset['asset']['browser_download_url']
        self.staticfile_manager = StaticFileUpdater(self.el, dwnl)
        fname, etag = self.staticfile_manager.download(status_update_cb)

        self.staticfile_manager = None
        return fname, target_asset['asset']['id']

    def cancel_download(self):
        if self.staticfile_manager:
            self.staticfile_manager.cancel_download()
            self.staticfile_manager = None

    def cleanup(self):
        if self.staticfile_manager:
            self.staticfile_manager.cleanup()

    def fetch_target_asset(self):
        update_data = None

        if self.embedded:
            update_data = self.get_embedded_data()
        else:
            config = self.get_config()
            repo = config.get('repo', '').split('/')

            if len(repo) < 2:
                return

            update_data = {
                'username': repo[0],
                'repo': repo[1], 
                'filename': config.get('repo_filename'),
            }
        
        if not update_data:
            return
        
        allow_prereleases = self.does_allow_prereleases()

        rel_url = '/'.join([
            'https://api.github.com/repos',
            update_data["username"],
            update_data["repo"],
            'releases',
        ])

        rel_data = []

        if not allow_prereleases:
            rel_url += f'/latest'

        try:
            rel_data_resp = requests.get(rel_url)
            rel_data_resp.raise_for_status()
            rel_data = rel_data_resp.json()
            if not allow_prereleases:
                rel_data = [rel_data]
        except Exception as e:
            if 'rate limit exceeded' in str(e):
                print(str(e))

            logging.error(e)
            return

        release = None

        possible_targets = []
        tmp_target_file = None

        for release in rel_data:
            found = False

            if not allow_prereleases and release['draft']:
                continue

            for asset in release['assets']:
                if fnmatch.fnmatch(asset['name'], update_data['filename']):
                    found = True
                    possible_targets.append(asset)
                    if self.embedded:
                        break

            if found:
                break

        if not release:
            return

        if len(possible_targets) == 1:
            tmp_target_file = possible_targets[0]
        else:
            logging.info(f'found {len(possible_targets)} possible file targets')
            
            for t in possible_targets:
                logging.info(' - ' + t['name'])

            # Check possible differences with system architecture in file name
            if self.system_arch == 'x86_64':
                for t in possible_targets:
                    if self.is_x86.search(t['name']) or not self.is_arm.search(t['name']):
                        tmp_target_file = t
                        logging.info('found possible target: ' + t['name'])
                        break

        if not tmp_target_file:
            logging.debug(f'No matching assets found from {rel_url}')
            return

        is_zsync = self.embedded and tmp_target_file['name'].endswith('.zsync')
        target_file = re.sub(r'\.zsync$', '', tmp_target_file['name'])

        for asset in release['assets']:
            if asset['name'] == target_file:
                logging.debug(f'Found 1 matching asset: {asset["name"]}')

                if is_zsync:
                    return {'asset': asset, 'zsync': tmp_target_file}

                return {'asset': asset, 'zsync': None}

    def is_update_available(self):
        if not self.el:
            return False

        target_asset = self.fetch_target_asset()

        if not os.path.exists(self.el.file_path):
            return False

        if target_asset:
            if target_asset['zsync']:
                logging.debug('GithubUpdated: checking zsync file at ' + target_asset['zsync']['browser_download_url'])
                zsync_file = requests.get(target_asset['zsync']['browser_download_url']).text
                zsync_file_header = zsync_file.split('\n\n', 1)[0]
                sha_pattern = r"SHA-1:\s*([0-9a-f]{40})"
                curr_version_hash = get_file_hash(Gio.File.new_for_path(self.el.file_path), alg='sha1')

                match = re.search(sha_pattern, zsync_file_header)
                if match:
                    return match.group(1) != curr_version_hash

            else:
                digest = target_asset['asset'].get('digest', '')
                if digest and digest.startswith('sha256:'):
                    curr_version_hash = get_file_hash(Gio.File.new_for_path(self.el.file_path), alg='sha256')
                    return f'sha256:{curr_version_hash}' != digest

                old_size = os.path.getsize(self.el.file_path)
                is_size_different = target_asset['asset']['size'] != old_size
                return is_size_different

        config = self.get_config()
        if config.get('repo_filename'):
            return None

        return False

    def load_form_rows(self): 
        config = self.get_config()
        repo = config.get('repo', '')
        filename = config.get('repo_filename', '')

        if self.embedded:
            url_data = self.get_embedded_data()
            if url_data:
                repo = '/'.join([url_data['username'], url_data['repo']])
                filename = url_data['filename']

        self.repo_url_row = AdwEntryRowDefault(
            icon_name='gl-git',
            sensitive=(not self.embedded),
            title=('Username/Repo')
        )

        self.repo_filename_row = AdwEntryRowDefault(
            icon_name='gl-paper',
            sensitive=(not self.embedded),
            title=_('Release file name')
        )

        if filename:
            self.repo_filename_row.set_text(filename)
        
        if repo:
            self.repo_url_row.set_text(repo)
        
        self.allow_prereleases_row = Adw.SwitchRow(
            title=_('Allow pre-releases'),
            sensitive=(not self.embedded),
            active=self.does_allow_prereleases()
        )

        return [
            self.repo_url_row, 
            self.repo_filename_row,
            self.allow_prereleases_row
        ]

    def get_config_from_form(self):
        allow_prereleases = False
        repo = None
        repo_filename = None

        if self.allow_prereleases_row:
            allow_prereleases = self.allow_prereleases_row.get_active()

        if self.repo_url_row:
            repo = self.repo_url_row.get_text().strip()

        if self.repo_filename_row:
            repo_filename = self.repo_filename_row.get_text().strip()

        return {
            'allow_prereleases': allow_prereleases,
            'repo': repo,
            'repo_filename': repo_filename,
        }

    def validate_config(self, config):
        if len(config.get('repo', '').split('/')) != 2:
            raise Exception(f'Invalid data, please enter <username>/<repo>')


from __future__ import annotations
from typing import Optional, Dict, List

from enum import Enum

class InstalledStatus(Enum):
    NOT_INSTALLED = 0
    INSTALLED = 1
    UNINSTALLING = 3
    ERROR = 4
    UNKNOWN = 5
    INSTALLING = 6
    UPDATE_AVAILABLE = 7
    UPDATING = 8

class AppListElement():
    def __init__(self, name: str, description: str, provider: str, installed_status: InstalledStatus, size: float=0):
        self.name: str = name
        self.description: str = description if description.strip() else _('No description provided')
        self.provider: str = provider
        self.installed_status: InstalledStatus = installed_status
        self.size: Optional[float] = size

    def set_installed_status(self, installed_status: InstalledStatus):
        self.installed_status = installed_status



import sys
import os
import logging
import json
import traceback
from .lib.constants import APP_ID
from gi.repository import Gio # noqa
from .BackgroudUpdatesFetcher import BackgroudUpdatesFetcher
from .lib.constants import FETCH_UPDATES_ARG
from .lib.utils import make_option, check_internet
from .providers.providers_list import appimage_provider
from .providers.AppImageProvider import AppImageUpdateLogic, AppImageListElement
from .lib.ini_config import Config
from .models.UpdateManagerChecker import UpdateManagerChecker
from .models.UpdateManager import UpdateManager

class Cli():
    options = [
        make_option('integrate', description='Integrate an AppImage file'),
        make_option('update', description='Update an AppImage file'),
        make_option('remove', description='Trashes an AppImage, its .desktop file and icons'),
        make_option('remove-all', description='Removes all AppImages'),
        make_option('list-installed', description='List integrated apps'),
        make_option('list-updates', description='List available updates'),
        make_option('list-update-managers', description='List available update managers'),
        make_option('set-update-url', description='(DEPRECATED, use set-update-source instead) Set a custom update url'),
        make_option('set-update-source', description='Set a custom update source'),
        make_option(FETCH_UPDATES_ARG, description='Fetch updates in the background and sends a desktop notification, used on system startup'),
    ]

    @staticmethod
    def ask(message: str, options: list) -> str:
        _input = None

        message = message.strip() + ' '

        while not _input in options:
            _input = input(message)

        return _input

    @staticmethod
    def from_options(argv):
        if len(argv) < 2:
            return -1

        opt = Cli._get_invoked_option(argv)

        if opt:
            name = str(opt.long_name).replace('-', '_')
            getattr(Cli, name)(argv)
            sys.exit(0)


        if '--help' in argv:
            table = [[f'--{o.long_name}', o.description] for o in Cli.options]
            text = f'Usage: gearlever [OPTION...]'
            text += f'\nFor a better user experience, please set `alias gearlever=\'gearlever\'` in your .bashrc file'
            Cli._print_help_if_requested(argv, table, text)

        return -1

    @staticmethod
    def fetch_updates(argv):
        BackgroudUpdatesFetcher.fetch()

    @staticmethod
    def update(argv):
        Cli._print_help_if_requested(argv, [
            ['--yes | -y', 'Skips any interactive question'],
            ['--force', 'Forces updates for running applications'],
            ['--all', 'Updates all AppImages'],
        ], text='Usage: --update <file_path>')

        assume_yes = ('-y' in argv) or ('--yes' in argv)
        update_all = ('--all' in argv)
        force = ('--force' in argv)

        updates: list[AppImageListElement] = []

        if not check_internet():
            print('Internet connection not available')
            sys.exit(1)

        if update_all:
            Cli.list_updates([])
            
            if not assume_yes:
                ans = Cli.ask('\nDo you really want to update all AppImages? (y/N)', ['y', 'Y', 'n', 'N'])
                if ans.lower() != 'y':
                    sys.exit(0)

            assume_yes = True
            installed = appimage_provider.list_installed()
            
            for el in installed:
                manager = UpdateManagerChecker.check_url_for_app(el)

                if manager and manager.is_update_available():
                    updates.append(el)
        else:
            g_file = Cli._get_file_from_args(argv)
            el = Cli._get_list_element_from_gfile(g_file)
            updates = [el]

        if not updates:
            sys.exit(0)

        for el in updates:
            if appimage_provider.is_app_running(el) and (not force):
                print(f'{el.file_path} was skipped because the application is running; use --force to skip this check.')
                continue

            manager = UpdateManagerChecker.check_url_for_app(el)

            if not manager:
                print('No update method was found for this AppImage')
                sys.exit(0)

            if not assume_yes:
                ans = Cli.ask('Do you really want to update this AppImage? (y/N)', ['y', 'Y', 'n', 'N'])
                if ans.lower() != 'y':
                    sys.exit(0)

            print(f'Downloading update from: {manager.name}')
            appimage_provider.update_from_url(manager, el, 
                lambda s: print("\rStatus: " + str(round(s * 100)) + "%", end=""))

            print(f'\n{el.file_path} updated successfully')

    @staticmethod
    def remove(argv):
        Cli._print_help_if_requested(argv, [
            ['Usage: --remove <file_path>'],
            ['--delete', 'Deletes the AppImage file from disk, instead of trashing it'],
            ['--yes | -y', 'Skips any interactive question'],
        ])

        g_file = Cli._get_file_from_args(argv)
        force = ('--delete' in argv)
        assume_yes = ('-y' in argv) or ('--yes' in argv)
        el = Cli._get_list_element_from_gfile(g_file)
        if not appimage_provider.is_installed(el):
            print('This AppImage is not integrated')
            sys.exit(1)

        q = 'Do you really want to delete this AppImage?'

        if force:
            q += ' This action is irreversible'

        ans = 'n'
        if not assume_yes:
            ans = Cli.ask(f'{q} (y/N)', ['y', 'Y', 'n', 'N'])

        if ans.lower() == 'y' or assume_yes:
            appimage_provider.uninstall(el, force_delete=force)
            print(f'{el.file_path} was removed sucessfully')

    @staticmethod
    def remove_all(argv):
        Cli._print_help_if_requested(argv, [
            ['Usage: --remove-all'],
            ['--yes | -y', 'Skips any interactive question'],
        ])

        assume_yes = ('-y' in argv) or ('--yes' in argv)
        apps = appimage_provider.list_installed()

        q = 'Do you really want to delete all AppImages?'

        ans = 'n'
        if not assume_yes:
            ans = Cli.ask(f'{q} (y/N)', ['y', 'Y', 'n', 'N'])

        if ans.lower() == 'y' or assume_yes:
            for el in apps:
                if appimage_provider.is_installed(el):
                    print(f'Removing {el.file_path}')
                    appimage_provider.uninstall(el, force_delete=True)

    @staticmethod
    def set_update_source(argv):
        u_managers = ', '.join([n.name for n in UpdateManagerChecker.get_models()])
        el = None

        manager_name = Cli._get_arg_value(argv, '--manager')
        selected_model: UpdateManager = UpdateManagerChecker.get_model_by_name(manager_name or '')
        manager_config_template = []

        if (not manager_name) or (not selected_model):
            Cli._print_help_if_requested(argv, [
                ['--manager <manager>', f'Specify an update manager between: {u_managers}'],
                ['--unset', f'Unset a custom config for an app'],
            ], text='Usage: --set-update-source <file_path> --manager <manager> [...OPTIONS]')

            print('Error: "%s" is not a valid update manager' % manager_name)
            sys.exit(1)

        manager: UpdateManager = selected_model(el=None)
        manager_config_template = manager.get_config_from_form()

        Cli._print_help_if_requested(argv, [
            [f'--manager {manager_name}'],
            *[[f'OPTION {k}=<value>'] for k in manager_config_template.keys()]
        ], text='Usage: --set-update-source <file_path> --manager <manager> [...OPTIONS]')

        g_file = Cli._get_file_from_args(argv)
        el = appimage_provider.create_list_element_from_file(g_file)
        manager: UpdateManager = selected_model(el=el, embedded=None) # type: ignore

        update_options = Cli._get_key_pairs(argv)

        if '--unset' in argv:
            Config.delete_app_update_config(el)
            sys.exit(0)

        if set(manager_config_template.keys()) != set(update_options.keys()):
            print('Missing or invalid update configuration, required keys: ' + ', '.join(manager_config_template.keys()))
            sys.exit(1)

        boolean_vals = ['0', '1', 'false', 'true']
        for k, v in manager_config_template.items():
            if type(v) == bool:
                if update_options[k] not in boolean_vals:
                    print(f'{k} is not a boolean value, allowed values: ' + '/'.join(boolean_vals))
                    sys.exit(1)

                update_options[k] = (update_options[k] in ['1', 'true'])

        manager.validate_config(update_options)

        Config.delete_app_update_config(el)
        Config.set_app_update_config(el, manager, update_options)

        sys.exit(0)

    @staticmethod
    def integrate(argv):
        Cli._print_help_if_requested(argv, [
            ['--keep-both', 'If a name conflict occurs, keeps both files (default behaviour)'],
            ['--replace', 'If a name conflict occurs, replaces the old file with the one that you are currently integrating'],
            ['--yes | -y', 'Skips any interactive question and integrates the file'],
            ['--update-url <url>', 'Set a custom URL for updates'],
        ], text='Usage: --integrate <file_path>')

        g_file = Cli._get_file_from_args(argv)

        list_element = appimage_provider.create_list_element_from_file(g_file)
        if appimage_provider.is_installed(list_element):
            print('This AppImage is already integrated')
            sys.exit(0)

        list_element.update_logic = AppImageUpdateLogic.KEEP
        appimage_provider.refresh_data(list_element)

        if '--replace' in argv:
            list_element.update_logic = AppImageUpdateLogic.REPLACE

        manager = UpdateManagerChecker.check_url_for_app(el=list_element)

        apps = appimage_provider.list_installed()

        already_installed = None
        for a in apps:
            if a.name == list_element.name:
                already_installed = a
                break

        if '--yes' not in argv \
            and '-y' not in argv:

            info_table = [
                ['Name', list_element.name,],
                ['Version', list_element.version or 'Not specified',],
                ['Description', list_element.description or 'None',],
                ['Update Source', 'None' if not manager else manager.name],
            ]

            Cli._print_table(info_table)
            ans = Cli.ask(f'\nDo you really want to integrate this AppImage? (y/N) ', ['y', 'Y', 'n', 'N'])

            if ans.lower() != 'y':
                sys.exit(0)

            if already_installed:
                ans = Cli.ask('\nAnother version of this app is already integrated.\nHow do you want to proceed? ((K)eep both/(R)eplace): ', 
                    ['k', 'r', 'K', 'R'])

            if ans.lower() == 'r':
                list_element.update_logic = AppImageUpdateLogic.REPLACE
                list_element.updating_from = already_installed

        appimage_provider.install_file(list_element)
        print(f'{list_element.file_path} was integrated successfully')

    @staticmethod
    def list_installed(argv):
        Cli._print_help_if_requested(argv, [])

        apps = appimage_provider.list_installed()

        table = []
        for a in apps:
            manager: UpdateManager | None = UpdateManagerChecker.check_url_for_app(a)
            update_mng = 'UpdatesNotAvailable'
            if manager:
                update_mng = manager.name

                if manager.embedded:
                    update_mng += '|embedded'

            update_mng = f'[{update_mng}]'

            v = a.version or 'Not specified'
            table.append([a.name, f'[{v}]', update_mng, a.file_path])

        Cli._print_table(table)

    @staticmethod
    def list_updates(argv):
        # Cli._print_help_if_requested(argv, [['-v', 'Prints update URL information']])

        installed = appimage_provider.list_installed()
        table = []

        if not check_internet():
            print('Internet connection not available')
            sys.exit(1)

        for el in installed:
            manager = UpdateManagerChecker.check_url_for_app(el)
            if not manager:
                continue

            try:
                if manager.is_update_available():
                    s = f'[Update available, {manager.name}]'
                    if manager.embedded:
                        s = f'[Update available, {manager.name}|embedded]'
                    row = [el.name, s, el.file_path]
                    # if '-v' in argv:
                    #     row.append(json.dumps(dict(manager.get_config())))

                    table.append(row)
            except Exception as e:
                logging.error(traceback.format_exc())

        if not table:
            print('No updates available')
            return

        Cli._print_table(table)

    @staticmethod
    def list_update_managers(argv):
        models = UpdateManagerChecker.get_models()
        table = []

        for m in models:
            table.append([m.label, m.name])

        Cli._print_table(table)

    @staticmethod
    def _get_key_pairs(argv: list[str]) -> dict:
        result = {}
        for pair in argv:
            if '=' in pair:
                key, value = pair.split('=', 1)
                result[key] = value

        return result

    @staticmethod
    def _get_arg_value(argv: list[str], arg_name: str):
        if not arg_name in argv:
            return None

        i = argv.index(arg_name)

        if len(argv) < (i + 2):
            print(f'Error: {arg_name} requires a value')
            sys.exit(1)

        val = argv[i + 1]
        return val

    @staticmethod
    def _print_table(table):
        if (not table):
            return

        longst_row = 0

        for r in table:
            if len(r) > longst_row:
                longst_row = len(r)

        
        for r in table:
            if len(r) < longst_row:
                while len(r) < longst_row:
                    r.append('')

        longest_cols = [
            (max([len(str(row[i])) for row in table]) + 3)
            for i in range(len(table[0]))
        ]

        row_format = "".join(["{:<" + str(longest_col) + "}" for longest_col in longest_cols])
        for row in table:
            print(row_format.format(*row))

    @staticmethod
    def _get_invoked_option(argv):
        for opt in Cli.options:
            long_name = str(opt.long_name)
            if f'--{long_name}' == argv[1]:
                return opt

        return None

    @staticmethod
    def _print_help_if_requested(argv, help: list, text=''):
        if '--help' in argv:
            opt = Cli._get_invoked_option(argv)
            if opt:
                print(str(opt.description))

            if text:
                print(text)

            Cli._print_table(help)
            print('\nMore documentation available at https://gearlever.mijorus.it')
            sys.exit(0)

    @staticmethod
    def _get_file_from_args(args):
        for a in args:
            if not a.startswith('-'):
                g_file = Gio.File.new_for_path(a)

                if os.path.exists(g_file.get_path()) \
                    and os.path.isfile(g_file.get_path()) \
                    and appimage_provider.can_install_file(g_file):
                    return g_file

        print('Error: please specify a valid AppImage file')
        sys.exit(1)

    @staticmethod
    def _get_list_element_from_gfile(g_file: Gio.File):
        el = None

        for a in appimage_provider.list_installed():
            if a.file_path == g_file.get_path():
                el = a
                break

        if not el:
            print('Error: AppImage not integrated')
            sys.exit(1)
    
        return el



# main.py
#
# Copyright 2022 Lorenzo Paderi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import gi
import logging
import shutil
import os

from .models.Settings import Settings
from .lib.ini_config import Config
from .lib import json_config
from .models.UpdateManagerChecker import UpdateManagerChecker
from .lib.constants import APP_ID, APP_NAME, APP_DATA, TMP_DIR
from .providers.providers_list import appimage_provider
from .GearleverWindow import GearleverWindow
from  .WelcomeScreen import WelcomeScreen
from .preferences import Preferences
from  .Cli import Cli

gi.require_version('Gtk', '4.0')
gi.require_version('Adw', '1')

from gi.repository import Gtk, Gio, Adw, Gdk, GLib, GObject # noqa

LOG_FILE_MAX_N_LINES = 5000
LOG_FOLDER = os.path.join(GLib.get_user_state_dir(), APP_NAME)
LOG_FILE = os.path.join(LOG_FOLDER, f'{APP_NAME}.log')

class GearleverApplication(Adw.Application):
    """The main application singleton class."""

    def __init__(self, version, pkgdatadir):
        super().__init__(application_id=APP_ID, flags=Gio.ApplicationFlags.HANDLES_OPEN)
        self.create_action('about', self.on_about_action)
        self.create_action('preferences', self.on_preferences_action, ['<primary>comma'])
        self.create_action('open_log_file', self.on_open_log_file)
        self.create_action('open_welcome_screen', self.on_open_welcome_screen, ['F1'])
        self.win = None
        self.version = version
        self.add_main_option_entries(Cli.options)

    def do_handle_local_options(self, options):
        return -1

    def do_startup(self):
        logging.info(f'\n\n---- Application startup | version {self.version}')
        Adw.Application.do_startup(self)

        settings = Settings.settings

        logging.debug('::: Settings')
        for k in settings.props.settings_schema.list_keys():
            logging.debug(k + ': ' + str(settings.get_value(k)))

        logging.debug('::: End settings')

        css_provider = Gtk.CssProvider()
        css_provider.load_from_resource(f'/it/mijorus/{APP_NAME}/assets/style.css')
        Gtk.StyleContext.add_provider_for_display(Gdk.Display.get_default(), css_provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)

    def do_activate(self, from_file=False):
        """Called when the application is activated.

        We raise the application's main window, creating it if
        necessary.
        """
        self.win = self.props.active_window

        if not self.win:
            self.win = GearleverWindow(application=self, from_file=from_file)
            self.win.connect('close-request', self.on_close_request)

        print('Logging to file ' + LOG_FILE)
        self.win.present()

    def do_open(self, files: list[Gio.File], n_files: int, data):
        if not files:
            return

        for f in files:
            if not appimage_provider.can_install_file(f):
                return
        
        self.do_activate(from_file=True)
        self.win.on_selected_local_file(list(files))

    def on_about_action(self, widget, data):
        about = Adw.AboutDialog(
            application_name='Gear Lever',
            version=self.version,
            developers=['Lorenzo Paderi'],
            copyright='2024 Lorenzo Paderi',
            application_icon='it.mijorus.gearlever',
            issue_url='https://github.com/mijorus/gearlever',
        )

        about.set_translator_credits(_("translator_credits"))
        if self.win:
            about.present(self.win)

    def on_preferences_action(self, widget, _):
        """Callback for the app.preferences action."""
        pref = Preferences()
        pref.connect('close-request', lambda w: self.win.on_show_installed_list() if self.win else None)
        pref.present()

    def create_action(self, name, callback, shortcuts=None):
        """Add an application action.

        Args:
            name: the name of the action
            callback: the function to be called when the action is
            activated
            shortcuts: an optional list of accelerators
        """
        action = Gio.SimpleAction.new(name, None)
        action.connect("activate", callback)
        self.add_action(action)
        if shortcuts:
            self.set_accels_for_action(f"app.{name}", shortcuts)

    def on_open_log_file(self, widget, event):
        if not self.win:
            return

        log_gfile = Gio.File.new_for_path(LOG_FOLDER)
        launcher = Gtk.FileLauncher.new(log_gfile)
        launcher.launch()

    def on_close_request(self, *args):
        if os.path.exists(TMP_DIR):
            shutil.rmtree(TMP_DIR)
        return False

    def on_open_welcome_screen(self, widget, event):
        tutorial = WelcomeScreen()
        if self.win:
            tutorial.present(self.win)

def migrate_to_v2_config():
    bg_update_k = 'fetch-updates-in-background'
    settings_conf = json_config.read_json_config('settings')
    Config.parser[Config.parser.default_section][bg_update_k] = Config.return_boolean(settings_conf.get(bg_update_k, False))
    installed_apps = appimage_provider.list_installed()

    for app in installed_apps:
        app_conf = json_config.read_config_for_app(app)
        Config.set_app_config(app, {
            'website': app_conf.get('website', '')
        })

        update_url_manager: str | None = app_conf.get('manager', None)

        if update_url_manager:
            model = None
            if update_url_manager:
                model = UpdateManagerChecker.get_model_by_name(update_url_manager)

            if model:
                update_manager = model(el=app)
                if update_manager:
                    update_manager.migrate_v2()

    Config.write()

def main(version, pkgdatadir):
    """The application's entry point."""
    APP_DATA['PKGDATADIR'] = pkgdatadir

    LOG_FILE = os.path.join(LOG_FOLDER, f'{APP_NAME}.log')

    if not os.path.exists(LOG_FOLDER):
        os.makedirs(LOG_FOLDER)

    if not os.path.exists(TMP_DIR):
        os.makedirs(TMP_DIR)

    # Clear log file if it's too big
    log_file_size = 0
    if os.path.exists(LOG_FILE): 
        with open(LOG_FILE, 'r') as f:
            log_file_size = len(f.readlines())
        
        if log_file_size > LOG_FILE_MAX_N_LINES:
            with open(LOG_FILE, 'w+') as f:
                f.write('')

    logging.basicConfig(
        filename=LOG_FILE,
        filemode='a',
        encoding='utf-8',
        format='%(asctime)s %(levelname)-1s [%(filename)s:%(lineno)d] %(message)s',
        level= logging.DEBUG if Settings.settings.get_boolean('debug-logs') else logging.INFO,
        force=True
    )

    if not Config.exists():
        migrate_to_v2_config()

    # old_configs_to_delete = ['settings', 'apps']
    # for name in old_configs_to_delete:
    #     p = os.path.join(GLib.get_user_config_dir(), f'{name}.json')
        
    #     if os.path.exists(p):
    #         os.remove(p)

    Config.refresh()

    if len(sys.argv) > 1:
        Cli.from_options(sys.argv)

    app = GearleverApplication(version, pkgdatadir)
    app.run(sys.argv)



from gi.repository import Gtk, Adw, GObject, Gio, GLib
from typing import Dict, List, Optional

import logging

from time import sleep
from .providers.providers_list import appimage_provider
from .providers.AppImageProvider import AppImageListElement
from .models.AppListElement import InstalledStatus
from .components.FilterEntry import FilterEntry
from .components.CustomComponents import NoAppsFoundRow
from .components.AppListBoxItem import AppListBoxItem
from .preferences import Preferences
from .WelcomeScreen import WelcomeScreen
from .lib.utils import get_application_window, check_internet
from .lib.async_utils import _async, idle
from .models.UpdateManagerChecker import UpdateManagerChecker
from .models.Settings import Settings

fetch_updates_cache = None

class InstalledAppsList(Gtk.ScrolledWindow):
    __gsignals__ = {
        "selected-app": (GObject.SIGNAL_RUN_FIRST, GObject.TYPE_NONE, (object, )),
        "update-all": (GObject.SIGNAL_RUN_FIRST, GObject.TYPE_NONE, (object, )),
    }

    CHECK_FOR_UPDATES_LABEL = _('Check for updates')
    NO_UPDATES_FOUND_LABEL = _('No updates found')
    CHECKING_FOR_UPDATES_LABEL = _('Checking updates...')
    NO_CONNECTION_LABEL = _('No Connection')
    UPDATE_ALL_LABEL = _('Update all')

    def __init__(self):
        super().__init__()
        self.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        clamp_size = 600

        self.container_stack = Gtk.Stack(transition_type=Gtk.StackTransitionType.CROSSFADE)
        self.main_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        self.updates_fetched = False

        self.installed_apps_list_slot = Gtk.Box()
        self.installed_apps_list = Gtk.ListBox(css_classes=["boxed-list"])
        self.installed_apps_list.connect('row-activated', self.on_activated_row)

        self.installed_apps_list_slot.append(self.installed_apps_list)

        self.installed_apps_list_rows: List[AppListBoxItem] = []
        self.no_apps_found_row = NoAppsFoundRow(visible=False)

        # Create the filter search bar
        self.filter_query: str = ''
        self.filter_entry = FilterEntry(_('Filter installed applications'), capture=get_application_window(), maximum_size=clamp_size)
        self.filter_entry.search_entry.connect('search-changed', self.trigger_filter_list)
        # self.filter_entry.set_search_mode(True)

        # title row
        title_row = Gtk.Box(margin_bottom=15, spacing=10)
        title_row.append( Gtk.Label(
            label=_('Installed applications'), 
            css_classes=['title-2'],
            hexpand=True,
            halign=Gtk.Align.START,
        ))

        # fetch updates btn
        self.updates_btn = Gtk.Button(
            label=self.CHECK_FOR_UPDATES_LABEL,
        )

        self.update_all_btn = Gtk.Button(
            label=self.UPDATE_ALL_LABEL,
            visible=False,
            css_classes=['suggested-action']
        )

        self.updates_btn.connect('clicked', self.on_fetch_updates_btn_clicked)
        self.update_all_btn.connect('clicked', self.on_update_all_btn_clicked)
        title_row.append(self.updates_btn)
        title_row.append(self.update_all_btn)

        [self.main_box.append(el) for el in [title_row, self.installed_apps_list_slot]]

        clamp = Adw.Clamp(child=self.main_box, maximum_size=clamp_size, margin_top=20, margin_bottom=20)

        self.clamp_container = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        [self.clamp_container.append(el) for el in [self.filter_entry, clamp]]

        # empty list placeholder
        builder = Gtk.Builder.new_from_resource('/it/mijorus/gearlever/gtk/empty-list-placeholder.ui')
        self.placeholder = builder.get_object('target')
        builder.get_object('open-preferences').connect('clicked', self.open_preferences)
        builder.get_object('show-welcome-screen').connect('clicked', self.on_open_welcome_screen)

        self.container_stack.add_child(self.clamp_container)
        self.container_stack.add_child(self.placeholder)

        self.set_child(self.container_stack)
        Settings.settings.connect('changed::appimages-default-folder', lambda *_: self.refresh_list())

    # Emit and event that changes the active page of the Stack in the parent widget
    def on_activated_row(self, listbox, row: Gtk.ListBoxRow):
        self.filter_entry.set_search_mode(False)
        self.emit('selected-app', row._app)

    def trigger_search_mode(self):
        self.filter_entry.set_search_mode(
            not self.filter_entry.get_search_mode()
        )

    def refresh_list(self):
        self.installed_apps_list.remove_all()
        self.updates_btn.set_label(self.CHECK_FOR_UPDATES_LABEL)
        self.installed_apps_list_rows = []

        installed: List[AppImageListElement] = appimage_provider.list_installed()

        for i in installed:
            list_row = AppListBoxItem(i, activatable=True, selectable=False, hexpand=True)
            list_row.set_update_version(i.version, i.size)

            list_row.load_icon()
            self.installed_apps_list_rows.append(list_row)
            self.installed_apps_list.append(list_row)

        if installed:
            self.container_stack.set_visible_child(self.clamp_container)
        else:
            self.container_stack.set_visible_child(self.placeholder)

        self.installed_apps_list.append(self.no_apps_found_row)
        self.no_apps_found_row.set_visible(False)
        
        self.installed_apps_list.set_sort_func(lambda r1, r2: self.sort_installed_apps_list(r1, r2))
        self.installed_apps_list.invalidate_sort()

        self.update_all_btn.set_visible(False)

    @_async
    def fetch_updates(self, cache=False):
        global fetch_updates_cache

        if not check_internet():
            def _show_no_connection():
                self.updates_btn.set_label(self.NO_CONNECTION_LABEL)
                self.updates_btn.add_css_class('destructive-action')
                self.updates_btn.set_sensitive(True)

            def _reset_btn():
                self.updates_btn.set_label(self.CHECK_FOR_UPDATES_LABEL)
                self.updates_btn.remove_css_class('destructive-action')
                return GLib.SOURCE_REMOVE

            GLib.idle_add(_show_no_connection)
            GLib.timeout_add_seconds(3, _reset_btn)
            return

        logging.debug('Fetching for updates for all apps')

        if cache and fetch_updates_cache:
            logging.debug('Getting updates list from cache')

            self.complete_updates_fetch(
                fetch_updates_cache['updatable_filepaths'], 
                fetch_updates_cache['updatable_apps'], 
                fetch_updates_cache['updates_available']
            )

            return

        GLib.idle_add(lambda: self.updates_btn.set_label(self.CHECKING_FOR_UPDATES_LABEL))
        GLib.idle_add(lambda: self.updates_btn.set_sensitive(False))

        updatable_apps = 0
        updates_available = 0
        final_rows = []
        for row in self.installed_apps_list_rows:
            manager = UpdateManagerChecker.check_url_for_app(row._app)

            updatable_apps += 1
            if not manager:
                continue

            logging.debug(f'Found app with update url: {manager.name}')

            try:
                status = manager.is_update_available()

                if status:
                    updates_available += 1
                    final_rows.append(row)
            except Exception as e:
                logging.error(e)

        self.updates_fetched = True
        updatable_filepaths = [f._app.file_path for f in final_rows]
        fetch_updates_cache = {
            'updatable_filepaths': updatable_filepaths, 
            'updatable_apps': updatable_apps, 
            'updates_available': updates_available
        }

        sleep(1)
        self.complete_updates_fetch(updatable_filepaths, updatable_apps, updates_available)

    @idle
    def complete_updates_fetch(self, updatable_filepaths: list[str], updatable_apps: int, updates_available: int):
        for row in self.installed_apps_list_rows:
            if row._app.file_path in updatable_filepaths:
                row.show_updatable_badge()

        self.update_all_btn.set_visible(updates_available > 0)
        if updates_available == 0:
            self.updates_btn.set_label(self.NO_UPDATES_FOUND_LABEL)
        else:
            self.updates_btn.set_icon_name('gl-arrow-circular-top-right')

        if updatable_apps:
            self.updates_btn.set_sensitive(True)

    def on_fetch_updates_btn_clicked(self, *args):
        self.update_all_btn.set_visible(False)
        self.fetch_updates()

    def trigger_filter_list(self, widget):
        if not self.installed_apps_list:
            return

        self.filter_query = widget.get_text()
        # self.installed_apps_list.invalidate_filter()

        for row in self.installed_apps_list_rows:
            if not getattr(row, 'force_show', False) and row._app.installed_status != InstalledStatus.INSTALLED:
                row.set_visible(False)
                continue

            if not len(self.filter_query):
                row.set_visible(True)
                continue

            visible = self.filter_query.lower().replace(' ', '') in row._app.name.lower()
            row.set_visible(visible)
            continue

        self.no_apps_found_row.set_visible(True)
        for row in self.installed_apps_list_rows:
            if row.get_visible():
                self.no_apps_found_row.set_visible(False)
                break

    def sort_installed_apps_list(self, row: AppListBoxItem, row1: AppListBoxItem):
        if (not hasattr(row1, '_app')):
            return 1

        if (not hasattr(row, '_app')) or (row._app.name.lower() < row1._app.name.lower()):
            return -1

        return 1

    def open_preferences(self, widget):
        pref = Preferences()
        pref.present()

    def on_update_all_btn_clicked(self, *args):
        self.emit('update-all', None)

    def on_open_welcome_screen(self, widget):
        tutorial = WelcomeScreen()
        tutorial.present(self)



import time
import logging
from .lib.terminal import sandbox_sh
from .lib.utils import check_internet
from .lib.constants import UPDATES_AVAILABLE_LABEL, ONE_UPDATE_AVAILABLE_LABEL, APP_ID
from .providers.AppImageProvider import AppImageProvider
from .models.UpdateManagerChecker import UpdateManagerChecker

class BackgroudUpdatesFetcher():
    @staticmethod
    def fetch():
        logging.warn('Fetching updates in the background')

        if not check_internet():
            return

        provider = AppImageProvider()
        installed = provider.list_installed()
        updates_available = 0

        for el in installed:
            manager = UpdateManagerChecker.check_url_for_app(el)

            if not manager:
                continue

            logging.debug(f'Checking app with: {manager.name}')

            try:
                status = manager.is_update_available()

                if status:
                    updates_available += 1
            except Exception as e:
                logging.error(e)

        if updates_available:
            content = ''
            if updates_available == 1:
                content = ONE_UPDATE_AVAILABLE_LABEL
            else:
                content = UPDATES_AVAILABLE_LABEL.replace('{n}', str(updates_available))

            expire_time = 1000 * 60 * 10
            open_label = _('Open') + ' Gear Lever'
            result = sandbox_sh(['notify-send', content, f'--action={open_label}', f'--expire-time={expire_time}'])
            if result == '0':
                sandbox_sh(['gearlever'])
        else:
            logging.warn('No available updates found')
            
        
    
    
    


import os
import random
from gi.repository import GLib # noqa

APP_ID = 'it.mijorus.gearlever'
APP_NAME = 'gearlever'
FETCH_UPDATES_ARG = 'fetch-updates'
APP_DATA = {
    'PKGDATADIR': ''
}

RUN_ID = ''.join((random.choice('abcdxyzpqr123456789') for i in range(10)))
ONE_UPDATE_AVAILABLE_LABEL = _('1 update available')
UPDATES_AVAILABLE_LABEL = _('{n} updates available')
TMP_DIR = os.path.join(GLib.get_tmp_dir(), APP_ID + '-' + RUN_ID)


import logging
from gi.repository import Gtk, GObject, Adw, Gdk, Gio, GLib

from .models.AppListElement import InstalledStatus
from .providers.AppImageProvider import AppImageListElement, AppImageUpdateLogic
from .providers.providers_list import appimage_provider
from .lib.async_utils import _async, idle, debounce
from .lib.utils import get_application_window
from .components.AppListBoxItem import AppListBoxItem

class MultiInstall(Gtk.ScrolledWindow):
    __gsignals__ = {
        "show-details": (GObject.SIGNAL_RUN_FIRST, GObject.TYPE_NONE, (object, )),
        "go-back": (GObject.SIGNAL_RUN_FIRST, GObject.TYPE_NONE, (bool, )),
    }

    def __init__(self) -> None:
        super().__init__()
        self.ACTION_ROW_ICON_SIZE = 45

        self.app_list: list[AppImageListElement] = []
        self.app_list_box_items: list[AppListBoxItem] = []
        self.app_list_box = Gtk.ListBox(css_classes=['boxed-list'])

        self.install_all_btn = Gtk.Button(
            css_classes=['suggested-action'], label=_('Move all to the app menu'), halign=Gtk.Align.CENTER)
        
        self.install_all_btn.connect('clicked', self.on_install_all_clicked)

        self.main_box = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL, 
            margin_top=30, 
            margin_bottom=10, 
            margin_start=20, 
            margin_end=20,
            spacing=20
        )

        self.main_box.append(self.install_all_btn)
        self.main_box.append(self.app_list_box)

        self.already_installed_warn = Gtk.Button(
            css_classes=['flat'],
            visible=False,
            sensitive=False,
            halign=Gtk.Align.CENTER,
            child=Adw.ButtonContent(
                label=_('Some apps are already installed'),
                icon_name='gl-info-symbolic'
            )
        )

        self.main_box.append(self.already_installed_warn)

        clamp = Adw.Clamp(child=self.main_box)
        overlay = Gtk.Overlay(
            child=clamp,
        )


        self.progress_bar = Gtk.ProgressBar(
            css_classes=['osd', 'horizontal', 'top'], fraction=1
        )

        overlay.add_overlay(self.progress_bar)
        overlay.set_clip_overlay(self.progress_bar, True)

        self.set_child(overlay)

    @idle
    def create_app_row_complete_load(self, el: AppImageListElement, icon: Gtk.Image):
        row = AppListBoxItem(el, show_details_btn=True)
        row.set_icon(icon)

        if el.installed_status is InstalledStatus.INSTALLED:
            row.set_opacity(0.5)

        row.details_btn.connect('clicked', self.on_details_btn_clicked, el)

        self.app_list_box.append(row)
        self.app_list_box_items.append(row)

        fraction = len(self.app_list_box_items) / len(self.app_list)
        self.progress_bar.set_fraction(fraction)

        if fraction == 1:
            self.progress_bar.set_visible(False)

            not_installed_count = self.count_not_installed()

            self.install_all_btn.set_sensitive(not_installed_count > 0)

    def count_not_installed(self):
        not_installed_count = 0
        for el in self.app_list:
            if el.installed_status is InstalledStatus.NOT_INSTALLED:
                not_installed_count += 1

        return not_installed_count

    @_async
    def create_app_row(self, el: AppImageListElement):
        appimage_provider.refresh_data(el)
        icon = appimage_provider.get_icon(el)

        self.create_app_row_complete_load(el, icon)

    def show_confirmation_dialog(self):
        dialog = Adw.MessageDialog(
            transient_for=get_application_window(),
            heading=_('Do you really want to move all the apps to the menu?')
        )

        dialog.add_response('cancel', _('Cancel'))
        dialog.add_response('confirm', _('Proceed'))
        dialog.set_response_appearance('confirm', Adw.ResponseAppearance.SUGGESTED)

        body = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        checkbox = Gtk.CheckButton.new_with_label(_('I have verified the source of the apps'))
        checkbox.set_halign(Gtk.Align.CENTER)

        body.append(checkbox)
        dialog.set_extra_child(body)
        dialog.connect('response', self.on_dialog_response, checkbox)

        dialog.present()

    def on_dialog_response(self, dialog: Adw.MessageDialog, response: str, checkbox: Gtk.CheckButton):
        if not checkbox.get_active():
            return
        
        if response != 'confirm':
            return
        
        self.install_all_btn.set_sensitive(False)

        for el in self.app_list:
            if el.installed_status is InstalledStatus.INSTALLED:
                continue

            el.set_trusted()
            appimage_provider.install_file(el)

        self.emit('go-back', True)

    def on_details_btn_clicked(self, widget: Gtk.Button, el: AppImageListElement):
        self.emit('show-details', el)

    @_async
    def create_list_elements(self, files):
        for f in files:
            try:
                el = appimage_provider.create_list_element_from_file(f)
                el.update_logic = AppImageUpdateLogic.KEEP
                self.app_list.append(el)

            except Exception as e:
                logging.error(e)

        self.progress_bar.set_visible(True)
        for el in self.app_list:
            self.create_app_row(el)

        not_installed_count = self.count_not_installed()
        show_warn = not_installed_count != len(self.app_list)
        
        GLib.idle_add(lambda: 
            self.already_installed_warn.set_visible(show_warn))

    def set_from_local_files(self, files: list[Gio.File]):
        if self.progress_bar.get_fraction() not in [0, 1]:
            return True

        self.progress_bar.pulse()
        self.app_list_box.remove_all()
        self.app_list = []
        self.app_list_box_items = []

        self.install_all_btn.set_sensitive(False)

        installable = 0
        for f in files:
            if appimage_provider.can_install_file(f):
                installable += 1

        self.create_list_elements(files)
        return installable > 0

    def on_install_all_clicked(self, widget: Gtk.Button):
        body = Gtk.BaselinePosition
        self.show_confirmation_dialog()


import subprocess
import re
import threading
from typing import Callable, List, Union, Optional
import logging
import os

def is_flatpak():
    return os.environ.get('FLATPAK_ID', False) != False

def host_sh(command: List[str], return_stderr=False, **kwargs) -> str:
    try:
        cmd = [*command]

        if is_flatpak():
            cmd = ['flatpak-spawn', '--host', *cmd]
        
        logging.debug(f'Running {cmd}')
        output = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, **kwargs)
        output.check_returncode()
    except subprocess.CalledProcessError as e:
        d = e.stderr.decode()
        if return_stderr:
            return d
        else:
            print(d)

        raise e

    output_string = output.stdout.decode()

    if return_stderr:
        output_string += output.stderr.decode()

    logging.debug(f'Done {cmd}')
    return re.sub(r'\n$', '', output_string)

def sandbox_sh(command: List[str], return_stderr=False, error_quiet=False, **kwargs) -> str:
    try:
        cmd = [*command]
        
        logging.debug(f'Running {command}')
        output = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, **kwargs)
        output.check_returncode()
    except subprocess.CalledProcessError as e:
        d = e.stderr.decode()
        if return_stderr:
            return d
        else:
            if not error_quiet:
                print(d)

        raise e

    output_string = output.stdout.decode()

    if return_stderr:
        output_string += output.stderr.decode()

    logging.debug(f'Done {cmd}')
    return re.sub(r'\n$', '', output_string)

def host_threaded_sh(command: List[str], callback: Optional[Callable[[str], None]]=None, return_stderr=False):
    def run_command(command: List[str], callback: Optional[Callable[[str], None]]=None):
        try:
            output = host_sh(command, return_stderr)

            if callback:
                callback(output)

        except subprocess.CalledProcessError as e:
            logging.error(e.stderr)
            raise e

    thread = threading.Thread(target=run_command, daemon=True, args=(command, callback, ))
    thread.start()


from gi.repository import Gtk, Adw, GObject, Gio, Gdk

class CenteringBox(Gtk.Box):
    def __init__(self, **kwargs):
        super().__init__(valign=Gtk.Align.CENTER, halign=Gtk.Align.CENTER, **kwargs)

class LabelStart(Gtk.Label):
    def __init__(self, **kwargs):
        super().__init__(halign=Gtk.Align.START, **kwargs)

class LabelCenter(Gtk.Label):
    def __init__(self, **kwargs):
        super().__init__(halign=Gtk.Align.CENTER, **kwargs)

class NoAppsFoundRow(Gtk.ListBoxRow):
    def __init__(self, **kwargs):
        super().__init__(hexpand=True)
        self.set_child(
            Gtk.Label(
                label="No apps found", 
                css_classes=['app-listbox-item'], 
                margin_bottom=20, 
                margin_top=20
            )
        )


import logging
import os
import ftputil
from ftputil.file import FTPFile
from ftputil import FTPHost
import fnmatch
import shutil
from typing import Optional
from urllib.parse import urlsplit
from ..lib.utils import get_random_string, get_file_hash
from ..lib import json_config
from ..lib.ini_config import Config

from .Models import DownloadInterruptedException
from .UpdateManager import UpdateManager
from .StaticFileUpdater import StaticFileUpdater
from ..components.AdwEntryRowDefault import AdwEntryRowDefault


# Example:
# https://download.kde.org/stable/digikam/

class FTPUpdater(UpdateManager):
    staticfile_manager: Optional[StaticFileUpdater]
    label = 'FTP'
    name = 'FTPUpdater'

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.staticfile_manager = None
        self.embedded = False

        self.url_row = None
        self.filename_row = None
        self.current_download: FTPHost | None = None

    def migrate_v2(self):
        app_config = json_config.read_config_for_app(self.el)
        config = None

        if 'update_manager_config' in app_config:
            old_config = app_config.get('update_manager_config', {})
            config = {
                'url': old_config.get('url', None),
                'filename': old_config.get('filename', None),
            }

        if config:
            Config.set_app_update_config(self.el, self, config)

    def download(self, status_update_cb) -> tuple[str, str]:
        conf = self.get_config()

        if not conf:
            raise Exception('Missing url data for FTPUpdater')

        target_asset = self.fetch_target_asset()
        if not target_asset:
            raise Exception('Missing target asset for FTPUpdater')

        random_name = get_random_string()

        if not os.path.exists(self.download_folder):
            os.makedirs(self.download_folder)

        fname = f'{self.download_folder}/{random_name}.appimage'

        server = conf['url'].replace('ftp://', '')

        self.current_download = ftputil.FTPHost(server, 'anonymous', '')
        chunk_size = 8192
        downloaded = 0
        oldperc = 0

        try:
            remote = self.current_download.open(target_asset['item_path'], 'rb')
            with open(fname, 'wb') as local:
                while True:
                    chunk = remote.read(chunk_size)

                    if not chunk:
                        break

                    local.write(chunk)
                    downloaded += len(chunk)
                    
                    # Calculate and display percentage
                    percent = (downloaded / target_asset['size'])
                    roundedperc = round(percent * 100)
                    if roundedperc != oldperc:
                        status_update_cb(percent)
                        oldperc = roundedperc

        except Exception as e:
            raise DownloadInterruptedException

        if self.current_download:
            self.current_download.close()
            self.current_download = None

        file_hash = get_file_hash(None, 'md5', file_path=fname)
        return fname, file_hash

    def cancel_download(self):
        if self.current_download:
            try:
                self.current_download.close()
            except Exception as e:
                pass
            finally:
                self.current_download = None

    def cleanup(self):
        if os.path.exists(self.download_folder):
            shutil.rmtree(self.download_folder)

    def fetch_target_asset(self):
        conf = self.get_config()
        pattern = conf['filename']
        matching_file = None

        server = conf['url'].replace('ftp://', '')
        with ftputil.FTPHost(server, 'anonymous', '') as ftp_host:
            # Parse the pattern to separate directory path from filename pattern
            parts = pattern.split('/')
            base_path = '/'
            wildcards_start = -1
            
            # Find where wildcards start
            for i, part in enumerate(parts):
                if '*' in part or '?' in part:
                    wildcards_start = i
                    break
            
            if wildcards_start > 0:
                base_path = '/'.join(parts[:wildcards_start])
            else:
                base_path = '/'.join(parts)
                if ftp_host.path.isfile(base_path):
                    size = ftp_host.path.getsize(base_path)
                    return {
                        'item_path': base_path, 
                        'size': size
                    }

                return None
            
            # Recursively find all matching files

            def find_matches(current_path, remaining_parts):
                """Recursively traverse directories to find matches."""
                if not remaining_parts:
                    return
                
                current_pattern = remaining_parts[0]
                
                try:
                    items = ftp_host.listdir(current_path)
                except:
                    return
                
                for item in items:
                    item_path = ftp_host.path.join(current_path, item)
                    logging.debug('found ' + str(item_path))
                    
                    # Check if item matches current pattern
                    if fnmatch.fnmatch(item, current_pattern):
                        if len(remaining_parts) == 1:
                            # Last pattern part - check if it's a file
                            if ftp_host.path.isfile(item_path):
                                logging.debug('FTPUpdater: Found mathing item ' + str(item_path))
                                size = ftp_host.path.getsize(item_path)

                                return {
                                    'item_path': item_path, 
                                    'size': size
                                }
                        else:
                            # More patterns to match - recurse into directory
                            if ftp_host.path.isdir(item_path):
                                return find_matches(item_path, remaining_parts[1:])
            
            # Start searching from base path
            pattern_parts = [p for p in parts[wildcards_start:] if p]
            matching_file = find_matches(base_path, pattern_parts)
            
        if not matching_file:
            logging.info("FTPUpdater: No files found matching the pattern")
            return None

        logging.debug(f'Found 1 matching asset: {matching_file["item_path"]}')
        return matching_file

    def is_update_available(self):
        target_asset = self.fetch_target_asset()

        if not target_asset:
            return False

        old_size = os.path.getsize(self.el.file_path)
        is_size_different = target_asset['size'] != old_size
        return is_size_different

    def load_form_rows(self, embedded=None):
        config = self.get_config()
        ftp_url = config.get('url')
        filename = config.get('filename')
        
        self.url_row = AdwEntryRowDefault(
            text=(ftp_url),
            icon_name='gl-earth-symbolic',
            sensitive=(not embedded),
            title=_('Server URL')
        )

        self.filename_row = AdwEntryRowDefault(
            text=filename,
            icon_name='gl-paper',
            sensitive=(not embedded),
            title=_('File name pattern')
        )

        return [self.url_row, self.filename_row]

    def get_config_from_form(self):
        url = None
        filename = None

        if self.url_row:
            url = self.url_row.get_text()

        if self.filename_row:
            filename = self.filename_row.get_text()

        return {
            'url': url,
            'filename': filename,
        }
    
    def validate_config(self, config):
        if not config.get('url', '').startswith('ftp://'):
            raise Exception(f'Invalid {self.name} url')





from typing import Dict
from .AppImageProvider import AppImageProvider

# A list containing all the "Providers"
# providers: Dict[str, Provider] = { 
#     'appimage': AppImageProvider()
# }

appimage_provider = AppImageProvider()


from typing import Optional, Callable
from gi.repository import Gtk, GObject, Adw
from ..lib.utils import get_application_window
from ..providers.AppImageProvider import AppImageListElement, AppImageUpdateLogic

class AppDetailsConflictModal():
    def __init__(self, app_name):
        self.modal = Adw.MessageDialog.new(
            get_application_window(), 
            _('Conflict with "{app_name}"').format(app_name=app_name), 
            _('There is already an app with the same name, how do you want to proceed?')
        )

        self.modal.add_response('cancel', _('Cancel'))
        self.modal.set_response_appearance('cancel', Adw.ResponseAppearance.DESTRUCTIVE)

        self.modal.add_response(AppImageUpdateLogic.KEEP.value, _('Keep both'))

        self.modal.add_response(AppImageUpdateLogic.REPLACE.value, _('Replace'))
        self.modal.set_response_appearance(AppImageUpdateLogic.REPLACE.value, Adw.ResponseAppearance.SUGGESTED)


import os
import platform
import re
import logging
from typing import Optional, Callable, Literal
from abc import ABC, abstractmethod

from ..lib.constants import TMP_DIR
from ..lib.ini_config import Config
from ..providers.AppImageProvider import AppImageListElement


class UpdateManager(ABC):
    name = ''
    label = ''
    handles_embedded: Optional[str] = None
    el: AppImageListElement = None
    system_arch = platform.machine()
    is_x86 = re.compile(r'(\-|\_|\.)x86(\-|\_|\.)')
    is_arm = re.compile(r'(\-|\_|\.)(arm64|aarch64|armv7l)(\-|\_|\.)')

    def __init__(self, el, embedded: Optional[str]=None) -> None:
        self.el = el
        self.download_folder = os.path.join(TMP_DIR, 'downloads')
        self.embedded = embedded

    def cleanup(self):
        pass

    @abstractmethod
    def is_update_available(self) -> 'bool | None':
        pass

    @abstractmethod
    def download(self, status_update_cb: Callable[[float], None]) -> tuple[str, str]:
        pass

    @abstractmethod
    def cancel_download(self):
        pass

    @abstractmethod
    def load_form_rows(self) -> list:
        pass

    @abstractmethod
    def get_config_from_form(self) -> dict:
        pass

    def get_config(self):
        return Config.get_app_update_config(self.el)
    
    def validate_config(self, config: dict):
        """Validates the configuration or raises an exception"""
        pass
    
    def migrate_v2(self):
        pass
